import mpi4py
mpi4py.rc.recv_mprobe = False

from mpi4py import MPI  # for libE communicator
import sys
import subprocess

task_nprocs = sys.argv[1]

size = MPI.COMM_WORLD.Get_size()
rank = MPI.COMM_WORLD.Get_rank()
runline = "mpirun -np ".split()
runline.append(task_nprocs)
runline.append('python')
runline.append('helloworld.py')

if rank == 0:
    print("Total sub-task procs: {}".format(size*int(task_nprocs)))
    print("Total procs (parent + sub-tasks): {}".format(size*(int(task_nprocs)+1)))

# print("Rank {}: {}".format(rank, " ".join(runline)))
output = 'task_' + str(rank) + '.out'
p = subprocess.Popen(runline, stdout=open(output, 'w'), shell=False)
p.wait()



#!/usr/bin/env python

"""
Plotting script for LibEnsemble run on WarpX simulations.
input: A directory which which the script will read the
latest .npy and .pickle files generated by LibEnsemble
"""

import os
import sys
import glob
import pickle
import numpy as np
import matplotlib.pyplot as plt

# 'aposmm' or 'random'
generator_type = 'random'

print("Reading latest .npy and .pickle files in " + sys.argv[1])

# Read latest .npy file, and store the result in results
# * means all if need specific format then *.csv
list_of_files = glob.glob(sys.argv[1] + '*npy')
latest_file = max(list_of_files, key=os.path.getctime)
results = np.load(latest_file)

# Read latest .pickle file, and store the result in pickle_data
# * means all if need specific format then *.csv
list_of_files = glob.glob(sys.argv[1] + '*pickle')
latest_file = max(list_of_files, key=os.path.getctime)
with open(latest_file, 'rb') as f:
    pickle_data = pickle.load(f)
if generator_type == 'aposmm':
    nbatches = len(pickle_data[1]['run_order'])

# Remove un-returned (i.e., not submitted) simulations
results = results[results['returned'] == 1]

# Re-organize results into a dictionary. Each key is a field in libE_output,
# each value is an array with the value of this field for all WarpX simulations
names = results.dtype.names
results_dict = {}
for i in range(len(names)):
    results_dict[names[i]] = np.array([task[i] for task in results])

# Sanity checks: exit if a simulation has null charge, as the script
# doesn't handle this (shouldn't be too hard to add though).
if np.min(results_dict['charge'] == 0.0):
    sys.exit()
if np.min(results_dict['f'] == 0.0):
    sys.exit()

# Custom: which output parameters are plotted
# as a function of which input parameters
plot_output = ['f', 'emittance', 'charge', 'energy_avg', 'energy_std']
plot_input = ['ramp_down_1', 'zlens_1', 'adjust_factor', 'ramp_down_2']

# For convenience, dictionary with pyplot.ylim
# tuples for all elements in plot_output
d_ylim = {
    'f': (3.e-7, 5.e-3),
    'emittance': (3.e-7, 5.e-3),
    'charge': (1.e-7, 1.e-5),
    'energy_avg': (5.e1, 5.e2),
    'energy_std': (1.e-2, 1.e-1)
}

# For convenience, dictionary with units for all elements in plot_output
d_yunits = {
    'f': " (a.u.)",
    'emittance': " (m rad)",
    'charge': " (pC/m)",
    'energy_avg': " (MeV)",
    'energy_std': " (dimless)"
}

# Print input and output parameters for the optimal run
print("Best run:")
ind_best = np.argmin(results_dict['f'])
for key in plot_input + plot_output:
    print(key, results_dict[key][ind_best])
print("charge_f/charge_i ",
      results[ind_best]['charge']/np.max(results_dict['charge']))

# And now plotting begins
plt.figure(figsize=(16, 10))
plt.style.use('dark_background')
# Loop over all input parameters in plot_input
for icount, iname in enumerate(plot_input):
    # Loop over all output parameters in plot_output
    for count, name in enumerate(plot_output):
        plt.subplot(len(plot_input), len(plot_output),
                    len(plot_output)*icount + count+1)
        if generator_type == 'aposmm':
            # Plot 1 point per run. Runs that are part of the same local
            # minimum search are batched together and plotted with the
            # same color.
            for batch in pickle_data[1]['run_order']:
                my_sims = np.isin(results_dict['sim_id'],
                                  pickle_data[1]['run_order'][batch])
                plt.scatter(np.squeeze(results_dict[iname][my_sims]),
                            np.squeeze(results_dict[name][my_sims]),
                            s=2)
        elif generator_type == 'random':
            # Plot 1 point per run. Colorbar stands for sim ID
            # (larger value = later run).
            plt.scatter(np.squeeze(results_dict[iname]),
                        np.squeeze(results_dict[name]),
                        c=results_dict['sim_id'], s=2, cmap='viridis')
            cbar = plt.colorbar()
            cbar.ax.set_ylabel('start time (arb. units)')
        else:
            print("generator_type should be either random or aposmm")
            sys.exit()
        plt.yscale('log')
        plt.xlim(np.min(results_dict[iname]), np.max(results_dict[iname]))
        plt.ylim(d_ylim[name])
        plt.grid()
        plt.title(name)
        plt.xlabel(iname)
        plt.ylabel(name + d_yunits[name])

plt.tight_layout()

# Save figure
plt.savefig('results.pdf', bbox_inches='tight')
plt.savefig('results.png', bbox_inches='tight')



import numpy as np
from libensemble.alloc_funcs.give_sim_work_first import give_sim_work_first
from libensemble.history import History


# -------------------------------------------------------------------------------------------------
# Set up sim_specs, gen_specs, exit_criteria

def make_criteria_and_specs_0(simx=10, n=1):
    sim_specs = {'sim_f': np.linalg.norm, 'in': ['x_on_cube'], 'persis_in': [],
                 'out': [('f', float), ('fvec', float, 3)], }
    gen_specs = {'gen_f': np.random.uniform,
                 'in': [],
                 'persis_in': [],
                 'out': [('priority', float), ('local_pt', bool), ('local_min', bool), ('num_active_runs', int)],
                 'user': {'ub': np.ones(n), 'lb': np.zeros(n), 'nu': 0}}
    if n == 1:
        gen_specs['out'] += [('x_on_cube', float)]
    else:
        gen_specs['out'] += [('x_on_cube', float, n)]
    exit_criteria = {'sim_max': simx}

    return sim_specs, gen_specs, exit_criteria


def make_criteria_and_specs_1(simx=10):
    sim_specs = {'sim_f': np.linalg.norm, 'in': ['x'], 'out': [('g', float)], }
    gen_specs = {'gen_f': np.random.uniform, 'in': [], 'out': [('x', float), ('priority', float)], 'user': {}}
    exit_criteria = {'sim_max': simx, 'stop_val': ('g', -1), 'elapsed_wallclock_time': 0.5}

    return sim_specs, gen_specs, exit_criteria


def make_criteria_and_specs_1A(simx=10):
    sim_specs = {'sim_f': np.linalg.norm, 'in': ['x'], 'out': [('g', float)], }
    gen_specs = {'gen_f': np.random.uniform,
                 'in': [],
                 'out': [('x', float), ('priority', float), ('sim_id', int)],
                 'user': {}}
    exit_criteria = {'sim_max': simx, 'stop_val': ('g', -1), 'elapsed_wallclock_time': 0.5}

    return sim_specs, gen_specs, exit_criteria


# -------------------------------------------------------------------------------------------------
# Set up history array
def hist_setup1(sim_max=10, n=1, H0_in=[]):
    sim_specs, gen_specs, exit_criteria = make_criteria_and_specs_0(simx=sim_max, n=n)
    alloc_specs = {'alloc_f': give_sim_work_first, 'out': [('allocated', bool)]}  # default for libE
    H0 = H0_in
    hist = History(alloc_specs, sim_specs, gen_specs, exit_criteria, H0)
    return hist, sim_specs, gen_specs, exit_criteria, alloc_specs


def hist_setup2(sim_max=10, H0_in=[]):
    sim_specs, gen_specs, exit_criteria = make_criteria_and_specs_1(simx=sim_max)
    alloc_specs = {'alloc_f': give_sim_work_first, 'out': [('allocated', bool)]}  # default for libE
    H0 = H0_in
    hist = History(alloc_specs, sim_specs, gen_specs, exit_criteria, H0)
    return hist, sim_specs, gen_specs, exit_criteria, alloc_specs


def hist_setup2A_genout_sim_ids(sim_max=10):
    sim_specs, gen_specs, exit_criteria = make_criteria_and_specs_1A(simx=sim_max)
    alloc_specs = {'alloc_f': give_sim_work_first, 'out': [('allocated', bool)]}  # default for libE
    H0 = []
    hist = History(alloc_specs, sim_specs, gen_specs, exit_criteria, H0)
    return hist, sim_specs, gen_specs, exit_criteria, alloc_specs



def rc(**kargs):
    """Runtime configuration options.

    Parameters
    ----------
    aposmm_optimizers : string or list of strings
        Select the aposmm optimizer/s (to prevent all options being imported).


    """
    for key in kargs:
        if not hasattr(rc, key):
            raise TypeError("unexpected argument '{0}'".format(key))
    for key, value in kargs.items():
        setattr(rc, key, value)


rc.aposmm_optimizers = None
__import__('sys').modules[__name__ + '.rc'] = rc



"""
The libE module is the outer libEnsemble routine.

This module sets up the manager and the team of workers, configured according
to the contents of the ``libE_specs`` dictionary. The manager/worker
communications scheme used in libEnsemble is parsed from the ``comms`` key
if present, with valid values being ``mpi``, ``local`` (for multiprocessing), or
``tcp``. MPI is the default; if a communicator is specified, each call to this
module will initiate manager/worker communications on a duplicate of that
communicator. Otherwise, a duplicate of COMM_WORLD will be used.

In the vast majority of cases, programming with libEnsemble involves the creation
of a *calling script*, a Python file where libEnsemble is parameterized via
the various specification dictionaries (e.g. :ref:`libE_specs<datastruct-libe-specs>`,
:ref:`sim_specs<datastruct-sim-specs>`, and :ref:`gen_specs<datastruct-gen-specs>`). The
outer libEnsemble routine ``libE()`` is imported and called with such dictionaries to initiate
libEnsemble. A simple calling script (from :doc:`the first tutorial<tutorials/local_sine_tutorial>`)
may resemble:

.. code-block:: python
    :linenos:

    import numpy as np
    from libensemble.libE import libE
    from generator import gen_random_sample
    from simulator import sim_find_sine
    from libensemble.tools import add_unique_random_streams

    nworkers, is_manager, libE_specs, _ = parse_args()

    libE_specs['save_every_k_gens'] = 20

    gen_specs = {'gen_f': gen_random_sample,
                 'out': [('x', float, (1,))],
                 'user': {
                    'lower': np.array([-3]),
                    'upper': np.array([3]),
                    'gen_batch_size': 5
                    }
                 }

    sim_specs = {'sim_f': sim_find_sine,
                 'in': ['x'],
                 'out': [('y', float)]}

    persis_info = add_unique_random_streams({}, nworkers+1)

    exit_criteria = {'sim_max': 80}

    H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info,
                                libE_specs=libE_specs)

This will initiate libEnsemble with a Manager and ``nworkers`` workers (parsed from
the command line), and runs on laptops or supercomputers. If an exception is
encountered by the manager or workers, the history array is dumped to file, and
MPI abort is called.

An alternative approach to parameterizing and interacting with libEnsemble via
``Ensemble`` objects and ``yaml`` files is available, but requires ``pyyaml``
to be installed. The equivalent of above resembles:

.. code-block:: python
    :linenos:

    import numpy as np
    from libensemble import Ensemble

    my_experiment = Ensemble()
    my_experiment.from_yaml('my_parameters.yaml')

    my_experiment.gen_specs['user']['lower'] = np.array([-3])
    my_experiment.gen_specs['user']['upper'] = np.array([3])

    H, persis_info, flag = my_experiment.run()

The remaining parameters may be found in a ``yaml`` file that resembles:

.. code-block:: yaml
    :linenos:

    libE_specs:
        save_every_k_gens: 20
        exit_criteria:
            sim_max: 80

    gen_specs:
        function: generator.gen_random_sample
        outputs:
            x:
                type: float
                size: 1
        user:
            gen_batch_size: 5

    sim_specs:
        function: simulator.sim_find_sine
        inputs:
            - x
        outputs:
            y:
                type: float


See below for the complete traditional ``libE()`` API.
"""

__all__ = ['libE']

import os
import logging
import random
import socket
import traceback
import numpy as np
import pickle  # Only used when saving output on error

from libensemble.version import __version__
from libensemble.utils import launcher
from libensemble.utils.timer import Timer
from libensemble.history import History
from libensemble.manager import manager_main, report_worker_exc, WorkerException, LoggedException
from libensemble.worker import worker_main
from libensemble.alloc_funcs import defaults as alloc_defaults
from libensemble.comms.comms import QCommProcess, Timeout
from libensemble.comms.logs import manager_logging_config
from libensemble.comms.tcp_mgr import ServerQCommManager, ClientQCommManager
from libensemble.executors.executor import Executor
from libensemble.resources.resources import Resources
from libensemble.tools.tools import _USER_SIM_ID_WARNING, osx_set_mp_method
from libensemble.tools.check_inputs import check_inputs

logger = logging.getLogger(__name__)
# To change logging level for just this module
# logger.setLevel(logging.DEBUG)


def libE(sim_specs, gen_specs, exit_criteria,
         persis_info=None,
         alloc_specs=None,
         libE_specs=None,
         H0=None):
    """
    Parameters
    ----------

    sim_specs: :obj:`dict`

        Specifications for the simulation function
        :doc:`(example)<data_structures/sim_specs>`

    gen_specs: :obj:`dict`

        Specifications for the generator function
        :doc:`(example)<data_structures/gen_specs>`

    exit_criteria: :obj:`dict`

        Tell libEnsemble when to stop a run
        :doc:`(example)<data_structures/exit_criteria>`

    persis_info: :obj:`dict`, optional

        Persistent information to be passed between user functions
        :doc:`(example)<data_structures/persis_info>`

    alloc_specs: :obj:`dict`, optional

        Specifications for the allocation function
        :doc:`(example)<data_structures/alloc_specs>`

    libE_specs: :obj:`dict`, optional

        Specifications for libEnsemble
        :doc:`(example)<data_structures/libE_specs>`

    H0: `NumPy structured array <https://docs.scipy.org/doc/numpy/user/basics.rec.html>`_, optional

        A previous libEnsemble history to be prepended to the history in the
        current libEnsemble run
        :doc:`(example)<data_structures/history_array>`

    Returns
    -------

    H: `NumPy structured array <https://docs.scipy.org/doc/numpy/user/basics.rec.html>`_

        History array storing rows for each point.
        :doc:`(example)<data_structures/history_array>`

    persis_info: :obj:`dict`

        Final state of persistent information
        :doc:`(example)<data_structures/persis_info>`

    exit_flag: :obj:`int`

        Flag containing final task status

        .. code-block::

            0 = No errors
            1 = Exception occurred
            2 = Manager timed out and ended simulation
            3 = Current process is not in libEnsemble MPI communicator
    """

    # Set default persis_info, alloc_specs, libE_specs, and H0
    if persis_info is None:
        persis_info = {}

    if alloc_specs is None:
        alloc_specs = alloc_defaults.alloc_specs

    if libE_specs is None:
        libE_specs = {}

    if H0 is None:
        H0 = np.empty(0)

    # Set default comms
    if 'comms' not in libE_specs:
        libE_specs['comms'] = 'mpi'

    libE_funcs = {'mpi': libE_mpi,
                  'tcp': libE_tcp,
                  'local': libE_local}

    comms_type = libE_specs.get('comms')

    assert comms_type in libE_funcs, "Unknown comms type: {}".format(comms_type)

    # Resource management not supported with TCP
    if comms_type == 'tcp':
        libE_specs['disable_resource_manager'] = True

    Resources.init_resources(libE_specs)

    return libE_funcs[comms_type](sim_specs, gen_specs, exit_criteria,
                                  persis_info, alloc_specs, libE_specs, H0)


def manager(wcomms, sim_specs, gen_specs, exit_criteria, persis_info,
            alloc_specs, libE_specs, hist,
            on_abort=None, on_cleanup=None):
    """Generic manager routine run."""

    logger.info('Logger initializing: [workerID] precedes each line. [0] = Manager')
    logger.info('libE version v{}'.format(__version__))

    if 'out' in gen_specs and ('sim_id', int) in gen_specs['out']:
        logger.manager_warning(_USER_SIM_ID_WARNING)

    save_H = libE_specs.get('save_H_and_persis_on_abort', True)

    try:
        try:
            persis_info, exit_flag, elapsed_time = \
                manager_main(hist, libE_specs, alloc_specs, sim_specs, gen_specs,
                             exit_criteria, persis_info, wcomms)
            logger.info("Manager total time: {}".format(elapsed_time))
        except LoggedException:
            # Exception already logged in manager
            raise
        except WorkerException as e:
            report_worker_exc(e)
            raise LoggedException(e.args[0], e.args[1]) from None
        except Exception as e:
            logger.error(traceback.format_exc())
            raise LoggedException(e.args) from None
    except Exception as e:
        exit_flag = 1  # Only exits if no abort/raise
        _dump_on_abort(hist, persis_info, save_H=save_H)
        if libE_specs.get('abort_on_exception', True) and on_abort is not None:
            on_abort()
        raise LoggedException(*e.args, 'See error details above and in ensemble.log') from None
    else:
        logger.debug("Manager exiting")
        logger.debug("Exiting with {} workers.".format(len(wcomms)))
        logger.debug("Exiting with exit criteria: {}".format(exit_criteria))
    finally:
        if on_cleanup is not None:
            on_cleanup()

    H = hist.trim_H()
    return H, persis_info, exit_flag


# ==================== MPI version =================================

class DupComm:
    """Duplicate MPI communicator for use with a with statement"""
    def __init__(self, mpi_comm):
        self.parent_comm = mpi_comm

    def __enter__(self):
        self.dup_comm = self.parent_comm.Dup()
        return self.dup_comm

    def __exit__(self, etype, value, traceback):
        self.dup_comm.Free()


def comms_abort(mpi_comm):
    """Abort all MPI ranks"""
    mpi_comm.Abort(1)  # Exit code 1 to represent an abort


def libE_mpi_defaults(libE_specs):
    """Fill in default values for MPI-based communicators."""

    from mpi4py import MPI

    if 'mpi_comm' not in libE_specs:
        libE_specs['mpi_comm'] = MPI.COMM_WORLD  # Will be duplicated immediately

    return libE_specs, MPI.COMM_NULL


def libE_mpi(sim_specs, gen_specs, exit_criteria,
             persis_info, alloc_specs, libE_specs, H0):
    """MPI version of the libE main routine"""

    libE_specs, mpi_comm_null = libE_mpi_defaults(libE_specs)

    if libE_specs['mpi_comm'] == mpi_comm_null:
        return [], persis_info, 3  # Process not in mpi_comm

    check_inputs(libE_specs, alloc_specs, sim_specs, gen_specs, exit_criteria, H0)

    with DupComm(libE_specs['mpi_comm']) as mpi_comm:
        rank = mpi_comm.Get_rank()
        is_manager = (rank == 0)

        resources = Resources.resources
        if resources is not None:
            local_host = socket.gethostname()
            libE_nodes = list(set(mpi_comm.allgather(local_host)))
            resources.add_comm_info(libE_nodes=libE_nodes)
            nworkers = mpi_comm.Get_size() - 1

        exctr = Executor.executor
        if exctr is not None:
            exctr.set_resources(resources)
            if is_manager:
                exctr.serial_setup()

        # Run manager or worker code, depending
        if is_manager:
            if resources is not None:
                resources.set_resource_manager(nworkers)
            return libE_mpi_manager(mpi_comm, sim_specs, gen_specs, exit_criteria,
                                    persis_info, alloc_specs, libE_specs, H0)

        # Worker returns a subset of MPI output
        libE_mpi_worker(mpi_comm, sim_specs, gen_specs, libE_specs)
        return [], {}, []


def libE_mpi_manager(mpi_comm, sim_specs, gen_specs, exit_criteria, persis_info,
                     alloc_specs, libE_specs, H0):
    """Manager routine runs on rank 0."""

    from libensemble.comms.mpi import MainMPIComm

    hist = History(alloc_specs, sim_specs, gen_specs, exit_criteria, H0)

    # Launch worker team
    wcomms = [MainMPIComm(mpi_comm, w) for w in
              range(1, mpi_comm.Get_size())]

    if not libE_specs.get('disable_log_files', False):
        manager_logging_config()

    # Set up abort handler
    def on_abort():
        """Shut down MPI on error."""
        comms_abort(mpi_comm)

    # Run generic manager
    return manager(wcomms, sim_specs, gen_specs, exit_criteria,
                   persis_info, alloc_specs, libE_specs, hist,
                   on_abort=on_abort)


def libE_mpi_worker(libE_comm, sim_specs, gen_specs, libE_specs):
    """Worker routines run on ranks > 0."""

    from libensemble.comms.mpi import MainMPIComm
    comm = MainMPIComm(libE_comm)
    worker_main(comm, sim_specs, gen_specs, libE_specs, log_comm=True)
    logger.debug("Worker {} exiting".format(libE_comm.Get_rank()))


# ==================== Local version ===============================


def start_proc_team(nworkers, sim_specs, gen_specs, libE_specs, log_comm=True):
    """Launch a process worker team."""
    wcomms = [QCommProcess(worker_main, sim_specs, gen_specs, libE_specs, w, log_comm)
              for w in range(1, nworkers+1)]
    for wcomm in wcomms:
        wcomm.run()
    return wcomms


def kill_proc_team(wcomms, timeout):
    """Join on workers (and terminate forcefully if needed)."""
    for wcomm in wcomms:
        try:
            wcomm.result(timeout=timeout)
        except Timeout:
            wcomm.terminate()


def libE_local(sim_specs, gen_specs, exit_criteria,
               persis_info, alloc_specs, libE_specs, H0):
    """Main routine for thread/process launch of libE."""

    nworkers = libE_specs['nworkers']

    check_inputs(libE_specs, alloc_specs, sim_specs, gen_specs, exit_criteria, H0)

    resources = Resources.resources
    if resources is not None:
        local_host = [socket.gethostname()]
        resources.add_comm_info(libE_nodes=local_host)

    exctr = Executor.executor
    if exctr is not None:
        exctr.set_resources(resources)
        exctr.serial_setup()

    hist = History(alloc_specs, sim_specs, gen_specs, exit_criteria, H0)

    # On Python 3.8 on macOS, the default start method for new processes was
    #  switched to 'spawn' by default due to 'fork' potentially causing crashes.
    # These crashes haven't yet been observed with libE, but with 'spawn' runs,
    #  warnings about leaked semaphore objects are displayed instead.
    # This function enforces 'fork' on macOS (Python 3.8)
    osx_set_mp_method()

    # Launch worker team and set up logger
    wcomms = start_proc_team(nworkers, sim_specs, gen_specs, libE_specs)

    # Set manager resources after the forkpoint.
    if resources is not None:
        resources.set_resource_manager(nworkers)

    if not libE_specs.get('disable_log_files', False):
        close_logs = manager_logging_config()

    # Set up cleanup routine to shut down worker team
    def cleanup():
        """Handler to clean up comms team."""
        kill_proc_team(wcomms, timeout=libE_specs.get('worker_timeout', 1))
        if close_logs is not None:  # logger remains set between multiple libE calls
            close_logs()

    # Run generic manager
    return manager(wcomms, sim_specs, gen_specs, exit_criteria,
                   persis_info, alloc_specs, libE_specs, hist,
                   on_cleanup=cleanup)


# ==================== TCP version =================================


def get_ip():
    """Get the IP address of the current host"""
    try:
        return socket.gethostbyname(socket.gethostname())
    except socket.gaierror:
        return 'localhost'


def libE_tcp_authkey():
    """Generate an authkey if not assigned by manager."""
    nonce = random.randrange(99999)
    return 'libE_auth_{}'.format(nonce)


def libE_tcp_default_ID():
    """Assign a (we hope unique) worker ID if not assigned by manager."""
    return "{}_pid{}".format(get_ip(), os.getpid())


def libE_tcp(sim_specs, gen_specs, exit_criteria,
             persis_info, alloc_specs, libE_specs, H0):
    """Main routine for TCP multiprocessing launch of libE."""

    check_inputs(libE_specs, alloc_specs, sim_specs, gen_specs, exit_criteria, H0)

    is_worker = True if 'workerID' in libE_specs else False

    exctr = Executor.executor
    if exctr is not None:
        # TCP does not currently support resource_management but when does, assume
        # each TCP worker is in a different resource pool (only knowing local_host)
        if not is_worker:
            exctr.serial_setup()

    if is_worker:
        libE_tcp_worker(sim_specs, gen_specs, libE_specs)
        return [], persis_info, []

    return libE_tcp_mgr(sim_specs, gen_specs, exit_criteria,
                        persis_info, alloc_specs, libE_specs, H0)


def libE_tcp_worker_launcher(libE_specs):
    """Get a launch function from libE_specs."""
    if 'worker_launcher' in libE_specs:
        worker_launcher = libE_specs['worker_launcher']
    else:
        worker_cmd = libE_specs['worker_cmd']

        def worker_launcher(specs):
            """Basic worker launch function."""
            return launcher.launch(worker_cmd, specs)
    return worker_launcher


def libE_tcp_start_team(manager, nworkers, workers,
                        ip, port, authkey, launchf):
    """Launch nworkers workers that attach back to a managers server."""
    worker_procs = []
    specs = {'manager_ip': ip, 'manager_port': port, 'authkey': authkey}
    with Timer() as timer:
        for w in range(1, nworkers+1):
            logger.info("Manager is launching worker {}".format(w))
            if workers is not None:
                specs['worker_ip'] = workers[w-1]
                specs['tunnel_port'] = 0x71BE
            specs['workerID'] = w
            worker_procs.append(launchf(specs))
        logger.info("Manager is awaiting {} workers".format(nworkers))
        wcomms = manager.await_workers(nworkers)
        logger.info("Manager connected to {} workers ({} s)".
                    format(nworkers, timer.elapsed))
    return worker_procs, wcomms


def libE_tcp_mgr(sim_specs, gen_specs, exit_criteria,
                 persis_info, alloc_specs, libE_specs, H0):
    """Main routine for TCP multiprocessing launch of libE at manager."""

    hist = History(alloc_specs, sim_specs, gen_specs, exit_criteria, H0)

    # Set up a worker launcher
    launchf = libE_tcp_worker_launcher(libE_specs)

    # Get worker launch parameters and fill in defaults for TCP/IP conn
    if 'nworkers' in libE_specs:
        workers = None
        nworkers = libE_specs['nworkers']
    elif 'workers' in libE_specs:
        workers = libE_specs['workers']
        nworkers = len(workers)
    ip = libE_specs.get('ip', None) or get_ip()
    port = libE_specs.get('port', 0)
    authkey = libE_specs.get('authkey', libE_tcp_authkey())

    osx_set_mp_method()

    with ServerQCommManager(port, authkey.encode('utf-8')) as tcp_manager:

        # Get port if needed because of auto-assignment
        if port == 0:
            _, port = tcp_manager.address

        if not libE_specs.get('disable_log_files', False):
            manager_logging_config()

        logger.info("Launched server at ({}, {})".format(ip, port))

        # Launch worker team and set up logger
        worker_procs, wcomms =\
            libE_tcp_start_team(tcp_manager, nworkers, workers,
                                ip, port, authkey, launchf)

        def cleanup():
            """Handler to clean up launched team."""
            for wp in worker_procs:
                launcher.cancel(wp, timeout=libE_specs.get('worker_timeout'))

        # Run generic manager
        return manager(wcomms, sim_specs, gen_specs, exit_criteria,
                       persis_info, alloc_specs, libE_specs, hist,
                       on_cleanup=cleanup)


def libE_tcp_worker(sim_specs, gen_specs, libE_specs):
    """Main routine for TCP worker launched by libE."""

    ip = libE_specs['ip']
    port = libE_specs['port']
    authkey = libE_specs['authkey']
    workerID = libE_specs['workerID']

    with ClientQCommManager(ip, port, authkey, workerID) as comm:
        worker_main(comm, sim_specs, gen_specs, libE_specs,
                    workerID=workerID, log_comm=True)
        logger.debug("Worker {} exiting".format(workerID))


# ==================== Additional Internal Functions ===========================


def _dump_on_abort(hist, persis_info, save_H=True):
    """Dump history and persis_info on abort"""
    logger.error("Manager exception raised .. aborting ensemble:")
    logger.error("Dumping ensemble history with {} sims evaluated:".
                 format(hist.returned_count))

    if save_H:
        np.save('libE_history_at_abort_' + str(hist.returned_count) + '.npy', hist.trim_H())
        with open('libE_persis_info_at_abort_' + str(hist.returned_count) + '.pickle', "wb") as f:
            pickle.dump(persis_info, f)



import os
import numpy as np
import pytest
import mock

from libensemble.libE import check_inputs, libE
from libensemble.manager import LoggedException
import libensemble.tests.unit_tests.setup as setup
from libensemble.alloc_funcs.give_sim_work_first import give_sim_work_first
from mpi4py import MPI
from libensemble.resources.resources import Resources
from libensemble.tests.regression_tests.common import mpi_comm_excl
from libensemble.comms.logs import LogConfig
from numpy import inf


class MPIAbortException(Exception):
    "Raised when mock mpi abort is called"


class MPISendException(Exception):
    "Raised when mock mpi abort is called"


class Fake_MPI:
    """Explicit mocking of MPI communicator"""

    def Get_size(self):
        return 2

    def Get_rank(self):
        return 0

    def Barrier(self):
        return 0

    def Dup(self):
        return self

    def Free(self):
        return

    def isend(self, msg, dest, tag):
        raise MPISendException()

    def Abort(self, flag):
        assert flag == 1, 'Aborting without exit code of 1'
        raise MPIAbortException()


fake_mpi = Fake_MPI()

alloc_specs = {'alloc_f': give_sim_work_first, 'out': [('allocated', bool)]}
hfile_abort = 'libE_history_at_abort_0.npy'
pfile_abort = 'libE_persis_info_at_abort_0.pickle'


# Run by pytest before each function
def setup_function(function):
    print("setup_function function:%s" % function.__name__)
    if Resources.resources is not None:
        del Resources.resources
        Resources.resources = None


def remove_file_if_exists(filename):
    try:
        os.remove(filename)
    except OSError:
        pass


def test_manager_exception():
    """Checking dump of history and pickle file on abort"""
    sim_specs, gen_specs, exit_criteria = setup.make_criteria_and_specs_0()
    remove_file_if_exists(hfile_abort)
    remove_file_if_exists(pfile_abort)

    with mock.patch('libensemble.manager.manager_main') as managerMock:
        managerMock.side_effect = Exception
        # Collision between libE.py and libE() (after mods to __init__.py) means
        #   libensemble.libE.comms_abort tries to refer to the function, not file
        with mock.patch('libensemble.comms_abort') as abortMock:
            abortMock.side_effect = Exception
            # Need fake MPI to get past the Manager only check and dump history
            with pytest.raises(Exception):
                libE_specs = {'mpi_comm': fake_mpi, 'disable_resource_manager': True}
                libE(sim_specs, gen_specs, exit_criteria, libE_specs=libE_specs)
                pytest.fail('Expected exception')
            assert os.path.isfile(hfile_abort), "History file not dumped"
            assert os.path.isfile(pfile_abort), "Pickle file not dumped"
            os.remove(hfile_abort)
            os.remove(pfile_abort)

            # Test that History and Pickle files NOT created when disabled
            with pytest.raises(Exception):
                libE_specs = {'mpi_comm': fake_mpi, 'save_H_and_persis_on_abort': False}
                libE(sim_specs, gen_specs, exit_criteria, libE_specs=libE_specs)
                pytest.fail('Expected exception')
            assert not os.path.isfile(hfile_abort), "History file dumped"
            assert not os.path.isfile(pfile_abort), "Pickle file dumped"


# Note - this could be combined now with above tests as fake_MPI prevents need for use of mock module
# Only way that is better is that this will simply hit first code exception - (when fake_MPI tries to isend)
# While first test triggers on call to manager
def test_exception_raising_manager_with_abort():
    """Running until fake_MPI tries to send msg to test (mocked) comm.Abort is called

    Manager should raise MPISendException when fakeMPI tries to send message, which
    will be caught by libE and raise MPIAbortException from fakeMPI.Abort"""
    with pytest.raises(MPIAbortException):
        sim_specs, gen_specs, exit_criteria = setup.make_criteria_and_specs_0()
        libE_specs = {'mpi_comm': fake_mpi, 'disable_resource_manager': True}
        libE(sim_specs, gen_specs, exit_criteria, libE_specs=libE_specs)
        pytest.fail('Expected MPIAbortException exception')


def test_exception_raising_manager_no_abort():
    """Running until fake_MPI tries to send msg to test (mocked) comm.Abort is called

    Manager should raise MPISendException when fakeMPI tries to send message, which
    will be caught by libE and raise MPIAbortException from fakeMPI.Abort"""
    libE_specs = {'abort_on_exception': False, 'mpi_comm': fake_mpi, 'disable_resource_manager': True}
    with pytest.raises(LoggedException):
        sim_specs, gen_specs, exit_criteria = setup.make_criteria_and_specs_0()
        libE(sim_specs, gen_specs, exit_criteria, libE_specs=libE_specs)
        pytest.fail('Expected MPISendException exception')


# So it's a key error rather than assertion error as does not test if 'in' is
# missing, only that it's a list - needs updating in future.
def test_exception_raising_check_inputs():
    """Intentionally running without sim_specs['in'] to test exception raising (Fails)"""
    libE_specs = {'mpi_comm': fake_mpi, 'disable_resource_manager': True}
    with pytest.raises(KeyError):
        H, _, _ = libE({'out': [('f', float)]}, {'out': [('x', float)]}, {'sim_max': 1},
                       libE_specs=libE_specs)
        pytest.fail('Expected KeyError exception')


def test_proc_not_in_communicator():
    """Checking proc not in communicator returns exit status of 3"""
    libE_specs = {}
    libE_specs['mpi_comm'], mpi_comm_null = mpi_comm_excl()
    H, _, flag = libE({'in': ['x'], 'out': [('f', float)]}, {'out': [('x', float)]},
                      {'sim_max': 1}, libE_specs=libE_specs)
    assert flag == 3, "libE return flag should be 3. Returned: " + str(flag)


# def test_exception_raising_worker():
#     # Intentionally running without sim_specs['in'] to test exception raising (Fails)
#     H, _, _ = libE({'out': [('f', float)]}, {'out': [('x', float)]},
#                    {'sim_max': 1}, libE_specs={'mpi_comm': MPI.COMM_WORLD})
#     assert H==[]


def rmfield(a, *fieldnames_to_remove):
    return a[[name for name in a.dtype.names if name not in fieldnames_to_remove]]


def check_assertion(libE_specs, alloc_specs, sim_specs, gen_specs, exit_criteria, H0):
    with pytest.raises(AssertionError) as excinfo:
        check_inputs(libE_specs, alloc_specs, sim_specs, gen_specs, exit_criteria, H0)
        pytest.fail('Expected AssertionError exception')
    return str(excinfo.value)


def test_checking_inputs_noworkers():
    # Don't take more points than there is space in history.
    sim_specs, gen_specs, exit_criteria = setup.make_criteria_and_specs_0()
    H0 = np.empty(0)
    # Should fail because only got a manager
    libE_specs = {'mpi_comm': MPI.COMM_WORLD, 'comms': 'mpi'}
    errstr = check_assertion(libE_specs, alloc_specs, sim_specs, gen_specs, exit_criteria, H0)
    assert 'must be at least one worker' in errstr, 'Incorrect assertion error: ' + errstr


def test_checking_inputs_H0():
    sim_specs, gen_specs, exit_criteria = setup.make_criteria_and_specs_0()
    libE_specs = {'mpi_comm': fake_mpi, 'comms': 'mpi'}

    # Should fail because H0 has points with 'return'==False
    H0 = np.array([(False, 0., 0, 0., 1, True, 1, True, [0., 0., 0.], True, 0.1, 1.1),
                   (False, 0., 0, 0., 1, True, 2, True, [0., 0., 0.], True, 0.2, 1.2),
                   (False, 0., 0, 0., 1, True, 3, True, [0., 0., 0.], True, 0.3, 1.3),
                   (False, 0., 0, 0., -1, False, 0, False, [0., 0., 0.], False, 0., inf),
                   (False, 0., 0, 0., -1, False, 0, False, [0., 0., 0.], False, 0., inf)],
                  dtype=[('local_pt', '?'), ('priority', '<f8'), ('gen_worker', '<i8'),
                         ('x_on_cube', '<f8'), ('sim_id', '<i8'), ('given', '?'),
                         ('sim_worker', '<i8'), ('returned', '?'), ('fvec', '<f8', (3,)),
                         ('allocated', '?'), ('f', '<f8'), ('given_time', '<f8')])

    # This should work
    check_inputs(libE_specs, alloc_specs, sim_specs, gen_specs, exit_criteria, H0)

    # A value has not been returned
    H0['returned'][2] = False
    errstr = check_assertion(libE_specs, alloc_specs, sim_specs, gen_specs, exit_criteria, H0)
    assert 'H0 contains unreturned or invalid points' in errstr, 'Incorrect assertion error: ' + errstr

    # Ungiven points shown as returned
    H0['returned'] = True
    errstr = check_assertion(libE_specs, alloc_specs, sim_specs, gen_specs, exit_criteria, H0)
    assert 'H0 contains unreturned or invalid points' in errstr, 'Incorrect assertion error: ' + errstr

    # Return to correct state
    H0['returned'][3:len(H0)] = False
    check_inputs(libE_specs, alloc_specs, sim_specs, gen_specs, exit_criteria, H0)

    # Removing 'returned' and then testing again. Should be successful as 'returned' does not exist
    H0 = rmfield(H0, 'returned')
    check_inputs(libE_specs, alloc_specs, sim_specs, gen_specs, exit_criteria, H0)

    # Should fail because H0 has fields not in H
    H0 = np.zeros(3, dtype=sim_specs['out'] + gen_specs['out'] + alloc_specs['out'] + [('bad_name2', bool)])
    errstr = check_assertion(libE_specs, alloc_specs, sim_specs, gen_specs, exit_criteria, H0)
    assert 'not in the History' in errstr, 'Incorrect assertion error: ' + errstr


def test_checking_inputs_exit_crit():
    sim_specs, gen_specs, _ = setup.make_criteria_and_specs_0()
    libE_specs = {'mpi_comm': fake_mpi, 'comms': 'mpi'}
    H0 = np.empty(0)

    exit_criteria = {}
    errstr = check_assertion(libE_specs, alloc_specs, sim_specs, gen_specs, exit_criteria, H0)
    assert 'Must have some exit criterion' in errstr, 'Incorrect assertion error: ' + errstr

    exit_criteria = {'swim_max': 10}
    errstr = check_assertion(libE_specs, alloc_specs, sim_specs, gen_specs, exit_criteria, H0)
    assert 'Valid termination options' in errstr, 'Incorrect assertion error: ' + errstr


def test_checking_inputs_single():
    sim_specs, gen_specs, exit_criteria = setup.make_criteria_and_specs_0()
    libE_specs = {'mpi_comm': fake_mpi, 'comms': 'mpi'}

    check_inputs(libE_specs=libE_specs)
    check_inputs(alloc_specs=alloc_specs)
    try:
        check_inputs(sim_specs=sim_specs)
    except AssertionError:
        assert 1, "Fails because sim_specs['in']=['x_on_cube'] and that's not an 'out' of anything"
    else:
        assert 0, "Should have failed"
    check_inputs(gen_specs=gen_specs)
    check_inputs(exit_criteria=exit_criteria, sim_specs=sim_specs, gen_specs=gen_specs)

    libE_specs['use_worker_dirs'] = True
    libE_specs['sim_input_dir'] = './__init__.py'
    libE_specs['sim_dir_copy_files'] = ['./__init__.py']
    check_inputs(libE_specs=libE_specs)


def test_logging_disabling():
    remove_file_if_exists('ensemble.log')
    remove_file_if_exists('libE_stats.txt')
    sim_specs, gen_specs, exit_criteria = setup.make_criteria_and_specs_0()
    libE_specs = {'mpi_comm': fake_mpi, 'comms': 'mpi', 'disable_log_files': True}
    logconfig = LogConfig.config
    logconfig.logger_set = False

    with mock.patch('libensemble.manager.manager_main') as managerMock:
        managerMock.side_effect = Exception
        with mock.patch('libensemble.comms_abort') as abortMock:
            abortMock.side_effect = Exception
            with pytest.raises(Exception):
                libE(sim_specs, gen_specs, exit_criteria, libE_specs=libE_specs)
                pytest.fail('Expected exception')
            assert not os.path.isfile('ensemble.log'), "ensemble.log file dumped"
            assert not os.path.isfile('libE_stats.txt'), "libE_stats.txt file dumped"


if __name__ == "__main__":
    test_manager_exception()
    test_exception_raising_manager_with_abort()
    test_exception_raising_manager_no_abort()
    test_exception_raising_check_inputs()
    test_proc_not_in_communicator()
    test_checking_inputs_noworkers()
    test_checking_inputs_H0()
    test_checking_inputs_exit_crit()
    test_checking_inputs_single()
    test_logging_disabling()



#!/usr/bin/env python

import sys
import numpy as np

if len(sys.argv) > 1:
    results = np.load(sys.argv[1])
else:
    print('You need to supply an .npy file - aborting')
    sys.exit()

done_only = False
dtype_only = False

if len(sys.argv) > 2:
    if sys.argv[2] == 'done':
        done_only = True
    elif sys.argv[2] == 'dtype':
        dtype_only = True
else:
    print(repr(results))

if done_only:
    count = 0
    for line in results:
        if line['returned']:
            count += 1

    results_filtered = np.zeros(count, dtype=results.dtype)
    count = 0
    for i, line in enumerate(results):
        if line['returned']:
            results_filtered[count] = results[i]
            count += 1

    print(repr(results_filtered))

elif dtype_only:
    print(results.dtype.names)
    print(type(results.dtype.names))



import numpy as np

from libensemble.message_numbers import STOP_TAG, PERSIS_STOP, FINISHED_PERSISTENT_GEN_TAG, EVAL_GEN_TAG
from libensemble.tools.persistent_support import PersistentSupport

__all__ = ['persistent_uniform',
           'uniform_random_sample_with_variable_resources',
           'persistent_request_shutdown']


def persistent_uniform(H, persis_info, gen_specs, libE_info):
    """
    This generation function always enters into persistent mode and returns
    ``gen_specs['initial_batch_size']`` uniformly sampled points the first time it
    is called. Afterwards, it returns the number of points given. This can be
    used in either a batch or asynchronous mode by adjusting the allocation
    function.

    .. seealso::
        `test_persistent_uniform_sampling.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_persistent_uniform_sampling.py>`_ # noqa
        `test_persistent_uniform_sampling_async.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_persistent_uniform_sampling_async.py>`_ # noqa
    """
    ub = gen_specs['user']['ub']
    lb = gen_specs['user']['lb']
    n = len(lb)
    b = gen_specs['user']['initial_batch_size']
    ps = PersistentSupport(libE_info, EVAL_GEN_TAG)

    # Send batches until manager sends stop tag
    tag = None
    while tag not in [STOP_TAG, PERSIS_STOP]:
        H_o = np.zeros(b, dtype=gen_specs['out'])
        H_o['x'] = persis_info['rand_stream'].uniform(lb, ub, (b, n))
        tag, Work, calc_in = ps.send_recv(H_o)
        if hasattr(calc_in, '__len__'):
            b = len(calc_in)

    H_o = None
    if gen_specs['user'].get('replace_final_fields', 0):
        # This is only to test libE ability to accept History after a
        # PERSIS_STOP. This history is returned in Work.
        H_o = Work
        H_o['x'] = -1.23

    return H_o, persis_info, FINISHED_PERSISTENT_GEN_TAG


def uniform_random_sample_with_variable_resources(H, persis_info, gen_specs, libE_info):
    """
    Generates points uniformly over the domain defined by ``gen_specs['user']['ub']`` and
    ``gen_specs['user']['lb']``. Also randomly requests a different number of resource
    sets to be used in the evaluation of the generated points after the initial batch.

    .. seealso::
        `test_uniform_sampling_with_variable_resources.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_uniform_sampling_with_variable_resources.py>`_ # noqa
    """
    ub = gen_specs['user']['ub']
    lb = gen_specs['user']['lb']
    n = len(lb)
    b = gen_specs['user']['initial_batch_size']
    ps = PersistentSupport(libE_info, EVAL_GEN_TAG)

    H_o = np.zeros(b, dtype=gen_specs['out'])
    for i in range(0, b):
        # x= i*np.ones(n)
        x = persis_info['rand_stream'].uniform(lb, ub, (1, n))
        H_o['x'][i] = x
        H_o['resource_sets'][i] = 1
        H_o['priority'] = 1

    # Send batches until manager sends stop tag
    tag, Work, calc_in = ps.send_recv(H_o)
    while tag not in [STOP_TAG, PERSIS_STOP]:
        H_o = np.zeros(b, dtype=gen_specs['out'])
        # H_o['x'] = len(H)*np.ones(n)
        H_o['x'] = persis_info['rand_stream'].uniform(lb, ub, (b, n))
        H_o['resource_sets'] = persis_info['rand_stream'].randint(1, gen_specs['user']['max_resource_sets']+1, b)
        H_o['priority'] = 10*H_o['resource_sets']
        print('Created {} sims, with worker_teams req. of size(s) {}'.format(b, H_o['resource_sets']), flush=True)
        tag, Work, calc_in = ps.send_recv(H_o)

        if calc_in is not None:
            b = len(calc_in)

    return H_o, persis_info, FINISHED_PERSISTENT_GEN_TAG


def persistent_request_shutdown(H, persis_info, gen_specs, libE_info):
    """
    This generation function is similar in structure to persistent_uniform,
    but uses a count to test exiting on a threshold value. This principle can
    be used with a supporting allocation function (e.g. start_only_persistent)
    to shutdown an ensemble when a condition is met.

    .. seealso::
        `test_persistent_uniform_gen_decides_stop.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_persistent_uniform_gen_decides_stop.py>`_ # noqa
    """
    ub = gen_specs['user']['ub']
    lb = gen_specs['user']['lb']
    n = len(lb)
    b = gen_specs['user']['initial_batch_size']
    shutdown_limit = gen_specs['user']['shutdown_limit']
    f_count = 0
    ps = PersistentSupport(libE_info, EVAL_GEN_TAG)

    # Send batches until manager sends stop tag
    tag = None
    while tag not in [STOP_TAG, PERSIS_STOP]:
        H_o = np.zeros(b, dtype=gen_specs['out'])
        H_o['x'] = persis_info['rand_stream'].uniform(lb, ub, (b, n))
        tag, Work, calc_in = ps.send_recv(H_o)
        if hasattr(calc_in, '__len__'):
            b = len(calc_in)
        f_count += b
        if f_count >= shutdown_limit:
            print('Reached threshold.', f_count, flush=True)
            break  # End the persistent gen

    return H_o, persis_info, FINISHED_PERSISTENT_GEN_TAG



# """
# Runs libEnsemble on a branin calculation with aposmm/nlopt.
#
# Execute via one of the following commands (e.g. 3 workers):
#    mpiexec -np 4 python3 test_branin_aposmm_nlopt_and_then_scipy.py
#
# The number of concurrent evaluations of the objective function will be 4-1=3.
# """

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local
# TESTSUITE_NPROCS: 3 4

import sys
import numpy as np
from copy import deepcopy
from pkg_resources import resource_filename
import shutil

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.sim_funcs.branin.branin_obj import call_branin as sim_f

import libensemble.gen_funcs
libensemble.gen_funcs.rc.aposmm_optimizers = ['nlopt', 'scipy']
from libensemble.gen_funcs.old_aposmm import aposmm_logic as gen_f

from libensemble.tests.regression_tests.support import (persis_info_2 as persis_info,
                                                        aposmm_gen_out as gen_out,
                                                        branin_vals_and_minima as M)
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams

nworkers, is_manager, libE_specs, _ = parse_args()

libE_specs['sim_input_dir'] = resource_filename('libensemble.sim_funcs.branin', '')  # to be copied by each worker

if libE_specs['comms'] == 'tcp':
    sys.exit("Cannot run with tcp when repeated calls to libE -- aborting...")

sim_specs = {'sim_f': sim_f,
             'in': ['x'],
             'out': [('f', float)]}

if nworkers == 3:
    sim_specs['user'] = {'uniform_random_pause_ub': 0.001}

n = 2
gen_out += [('x', float, n), ('x_on_cube', float, n)]
gen_specs = {'gen_f': gen_f,
             'in': [o[0] for o in gen_out] + ['f', 'returned'],
             'out': gen_out,
             'user': {'lb': np.array([-5, 0]),
                      'ub': np.array([10, 15]),
                      'initial_sample_size': 20,
                      'localopt_method': 'LN_BOBYQA',
                      'dist_to_bound_multiple': 0.99,
                      'xtol_rel': 1e-3,
                      'min_batch_size': nworkers,
                      'high_priority_to_best_localopt_runs': True,
                      'max_active_runs': 3}
             }

persis_info = add_unique_random_streams(persis_info, nworkers + 1)
persis_info_safe = deepcopy(persis_info)

# Tell libEnsemble when to stop (stop_val key must be in H)
exit_criteria = {'sim_max': 150,
                 'stop_val': ('f', -1)}
# end_exit_criteria_rst_tag

# Perform the run
for run in range(2):
    libE_specs['ensemble_dir_path'] = './ensemble_w' + str(nworkers) + '_r' + str(run)
    if run == 1:
        gen_specs['user']['localopt_method'] = 'scipy_COBYLA'
        gen_specs['user'].pop('xtol_rel')
        gen_specs['user']['tol'] = 1e-5
        exit_criteria['sim_max'] = 500
        persis_info = deepcopy(persis_info_safe)

    H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria,
                                persis_info, libE_specs=libE_specs)

    if is_manager:
        M = M[M[:, -1].argsort()]  # Sort by function values (last column)
        k = M.shape[0]
        tol = 1e-5
        for i in range(k):
            dist = np.min(np.sum((H['x'][H['local_min']]-M[i, :2])**2, 1))
            print(dist)
            assert dist < tol

        print("\nAPOSMM + " + gen_specs['user']['localopt_method'] +
              " found " + str(k) + " minima to tolerance " + str(tol))
        save_libE_output(H, persis_info, __file__, nworkers)

        shutil.rmtree(libE_specs['ensemble_dir_path'])



# """
# Runs libEnsemble on the 6-hump camel problem. Documented here:
#    https://www.sfu.ca/~ssurjano/camel6.html
#
# Execute via one of the following commands (e.g. 3 workers):
#    mpiexec -np 4 python3 test_6-hump_camel_aposmm_LD_MMA.py
#
# The number of concurrent evaluations of the objective function will be 4-1=3.
# """

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi
# TESTSUITE_NPROCS: 2 4

import sys
import numpy as np
from math import gamma, pi, sqrt
from copy import deepcopy

# Import libEnsemble items for this test
from libensemble.libE import libE, libE_tcp_worker
from libensemble.sim_funcs.six_hump_camel import six_hump_camel as sim_f

import libensemble.gen_funcs
libensemble.gen_funcs.rc.aposmm_optimizers = ['nlopt', 'petsc']
from libensemble.gen_funcs.old_aposmm import aposmm_logic as gen_f

from libensemble.alloc_funcs.fast_alloc_to_aposmm import give_sim_work_first as alloc_f
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams
from libensemble.tests.regression_tests.support import (persis_info_1 as persis_info,
                                                        aposmm_gen_out as gen_out,
                                                        six_hump_camel_minima as minima)

nworkers, is_manager, libE_specs, _ = parse_args()

n = 2
sim_specs = {'sim_f': sim_f,
             'in': ['x'],
             'out': [('f', float), ('grad', float, n)]}

gen_out += [('x', float, n), ('x_on_cube', float, n)]
gen_specs = {'gen_f': gen_f,
             'in': [o[0] for o in gen_out] + ['f', 'grad', 'returned'],
             'out': gen_out,
             'user': {'initial_sample_size': 100,
                      'sample_points': np.round(minima, 1),
                      'localopt_method': 'LD_MMA',
                      'rk_const': 0.5*((gamma(1+(n/2))*5)**(1/n))/sqrt(pi),
                      'xtol_rel': 1e-3,
                      'max_active_runs': 6,
                      'lb': np.array([-3, -2]),
                      'ub': np.array([3, 2])
                      }
             }

alloc_specs = {'alloc_f': alloc_f, 'out': [('allocated', bool)], 'user': {'batch_mode': True, 'num_active_gens': 1}}

persis_info = add_unique_random_streams(persis_info, nworkers + 1)
persis_info_safe = deepcopy(persis_info)

exit_criteria = {'sim_max': 1000}

# Set up appropriate abort mechanism depending on comms
libE_abort = quit
if libE_specs['comms'] == 'mpi':
    from mpi4py import MPI

    def libE_mpi_abort():
        MPI.COMM_WORLD.Abort(1)

    libE_abort = libE_mpi_abort

# Perform the run (TCP worker mode)
if libE_specs['comms'] == 'tcp' and not is_manager:
    run = int(sys.argv[-1])
    libE_tcp_worker(sim_specs, gen_specs[run], libE_specs)
    quit()

# Perform the run
for run in range(3):
    if libE_specs['comms'] == 'tcp' and is_manager:
        libE_specs['worker_cmd'].append(str(run))

    if run == 1:
        gen_specs['user']['localopt_method'] = 'blmvm'
        gen_specs['user']['grtol'] = 1e-5
        gen_specs['user']['gatol'] = 1e-5
        persis_info = deepcopy(persis_info_safe)

    if run == 2:
        gen_specs['user']['localopt_method'] = 'LD_MMA'
        # Change the bounds to put a local min at a corner point (to test that
        # APOSMM handles the same point being in multiple runs) ability to
        # give back a previously evaluated point)
        gen_specs['user']['ub'] = np.array([-2.9, -1.9])
        gen_specs['user']['mu'] = 1e-4
        gen_specs['user']['rk_const'] = 0.01*((gamma(1+(n/2))*5)**(1/n))/sqrt(pi)
        gen_specs['user']['lhs_divisions'] = 2
        # APOSMM can be called when some run is incomplete
        alloc_specs['user'].pop('batch_mode')

        gen_specs['user'].pop('xtol_rel')
        gen_specs['user']['ftol_rel'] = 1e-2
        gen_specs['user']['xtol_abs'] = 1e-3
        gen_specs['user']['ftol_abs'] = 1e-8
        exit_criteria = {'sim_max': 200}
        minima = np.array([[-2.9, -1.9]])

        persis_info = deepcopy(persis_info_safe)

    H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria,
                                persis_info, alloc_specs, libE_specs)

    if is_manager:
        if flag != 0:
            print("Exit was not on convergence (code {})".format(flag), flush=True)
            libE_abort()

        tol = 1e-5
        for m in minima:
            # The minima are known on this test problem.
            # 1) We use their values to test APOSMM has identified all minima
            # 2) We use their approximate values to ensure APOSMM evaluates a
            #    point in each minima's basin of attraction.
            print(np.min(np.sum((H[H['local_min']]['x'] - m)**2, 1)), flush=True)
            if np.min(np.sum((H[H['local_min']]['x'] - m)**2, 1)) > tol:
                libE_abort()

        print("\nlibEnsemble with APOSMM using a gradient-based localopt method has identified the " +
              str(np.shape(minima)[0]) + " minima within a tolerance " + str(tol))
        save_libE_output(H, persis_info, __file__, nworkers)



"""
Tests libEnsemble n-agent distributed optimization generator function on a
variety of cases, including:
   - Rosenbrock function
   - Alternative Rosenbrock function
   - Nesterov's quadratic function (Sec 2.1.2 in Nesterov's 's "Introductory Lectures on Convex Programming")
   - Linear regression with l2 regularization
   - Logistic regression with l2 regularization
   - PYCUTEst functions (https://jfowkes.github.io/pycutest/_build/html/index.html)

You can specify which problem to test by setting @prob_id in {0,1,2,3,4,5}.

This call script uses distributed gradient-tracking (https://doi.org/10.1109/TCNS.2020.3024321)
to solve the following problems. To test, run using, for any p >= 6,
   mpiexec -np p python3 test_persistent_n_agent.py
   python3 test_persistent_n_agent.py --nworkers p --comms local

The number gens will be 4.

Note that this library makes use the NLOPT library to obtain the optimal value
for regression testing. To test PYCUTEst, make sure to install the necessary
files. Refer to tools/pycute_interface on more details. To ignore this library,
simply comment out the import of "Blackbox" below
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local
# TESTSUITE_NPROCS: 6
# TESTSUITE_OS_SKIP: OSX

import numpy as np
import numpy.linalg as la
import scipy.sparse as spp

import sys
from libensemble.libE import libE
from libensemble.gen_funcs.persistent_n_agent import n_agent as gen_f
from libensemble.alloc_funcs.start_persistent_consensus import start_consensus_persistent_gens as alloc_f
from libensemble.tools import parse_args, add_unique_random_streams
from libensemble.tools.consensus_subroutines import get_k_reach_chain_matrix, get_doubly_stochastic, regls_opt, log_opt

from libensemble.sim_funcs.rosenbrock import rosenbrock_eval
from libensemble.sim_funcs.alt_rosenbrock import alt_rosenbrock_eval
from libensemble.sim_funcs.nesterov_quadratic import nesterov_quadratic_eval
from libensemble.sim_funcs.linear_regression import linear_regression_eval
from libensemble.sim_funcs.logistic_regression import logistic_regression_eval

nworkers, is_manager, libE_specs, _ = parse_args()

if nworkers < 2:
    sys.exit("Cannot run with a persistent worker if only one worker -- aborting...")
if nworkers < 5:
    sys.exit('This tests requires at least 5 workers (6 MPI processes)...')

num_gens = 4
A = spp.diags([2, 3, 3, 2]) - get_k_reach_chain_matrix(num_gens, 2)
S = get_doubly_stochastic(A)
rho = la.norm(S - (S.shape[0] ** -1) * np.ones(S.shape), ord=2)

eps = 1e-1
N_const = 100

# Perform the run
libE_specs['safe_mode'] = False

# 0: rosenbrock, 1: alt rosenbrock, 2: nesterov's, 3: l2 linear regression
# 4: l2 logistic regression, 5: l2 logistic w/ early stop
for prob_id in range(6):
    persis_info = {}
    persis_info['print_progress'] = 0
    persis_info['A'] = S
    persis_info = add_unique_random_streams(persis_info, nworkers + 1)
    persis_info['gen_params'] = {}

    if prob_id <= 4:
        exit_criteria = {'elapsed_wallclock_time': 600}
    else:
        exit_criteria = {'sim_max': 50}

    if prob_id == 0:
        sim_f = rosenbrock_eval
        m, n = 10, 20
        prob_name = 'Chained Rosenbrock'
        L = 1
        fstar = 0
        err_const = 1e3

    if prob_id == 1:
        sim_f = alt_rosenbrock_eval
        m, n = 14, 15
        prob_name = 'Alternative chained Rosenbrock'
        L = 1
        fstar = 0
        err_const = 1e3

    if prob_id == 2:
        sim_f = nesterov_quadratic_eval
        m, n = 15, 14
        prob_name = "Nesterov's quadratic function"
        L = 4
        # See Sec 2.1.2 of Nesterov's "Introductory Lectures on Convex Programming"
        fstar = 0.5 * (-1 + 1 / (m + 1))
        err_const = 1

    if prob_id == 3:
        sim_f = linear_regression_eval
        m, n = 4, 15
        prob_name = 'linear regression with l2 regularization'
        L = 1
        err_const = 1e1
        X = np.array([np.random.normal(loc=0, scale=1.0, size=n) for _ in range(m)]).T
        y = np.dot(X.T, np.ones(n)) + np.cos(np.dot(X.T, np.ones(n))) + np.random.normal(loc=0, scale=0.25, size=m)
        c = 0.1

        X_norms = la.norm(X, ord=2, axis=0) ** 2
        L = (2 / m) * (np.amax(X_norms) + c)

        # reduce size of problem to match available gens
        persis_info['sim_params'] = {'X': X, 'y': y, 'c': c, 'reg': 'l2'}
        fstar = regls_opt(X, y, c, reg='l2')

        def df(theta, i):
            return (2 / m) * (-y[i] + np.dot(X[:, i], theta)) * X[:, i] + (2 * c / m) * theta

        def f(theta, i):
            z = y[i] - np.dot(X[:, i], theta)
            return (1 / m) * np.dot(z, z) + (c / m) * np.dot(theta, theta)

        # Setting @f_i_eval and @df_i_eval tells to gen to compute gradients locally
        persis_info['gen_params'] = {'f_i_eval': f, 'df_i_eval': df}

    if prob_id >= 4:
        sim_f = logistic_regression_eval
        m, n = 14, 15
        prob_name = 'logistic regression  with l2 regularization'
        if prob_id > 4:
            prob_name += ' w/ stoppage'
        L = 1
        err_const = 1e1
        y = np.append(2 * np.ones(m // 2), np.zeros(m - m // 2)) - 1
        X = np.array([np.random.normal(loc=y[i] * np.ones(n), scale=1.0, size=n) for i in range(m)]).T
        c = 0.1

        XXT_sum = np.outer(X[:, 0], X[:, 0])
        for i in range(1, m):
            XXT_sum += np.outer(X[:, i], X[:, i])
        eig_max = np.amax(la.eig(XXT_sum)[0].real)
        L = eig_max / m

        persis_info['sim_params'] = {'X': X, 'y': y, 'c': c, 'reg': 'l2'}
        fstar = log_opt(X, y, c, 'l2')

    sim_specs = {
        'sim_f': sim_f,
        'in': ['x', 'obj_component', 'get_grad'],
        'out': [('f_i', float), ('gradf_i', float, (n,))],
    }

    gen_specs = {
        'gen_f': gen_f,
        'out': [
            ('x', float, (n,)),
            ('f_i', float),
            ('eval_pt', bool),  # eval point
            ('consensus_pt', bool),  # does not require a sim
            ('obj_component', int),  # which {f_i} to eval
            ('get_grad', bool),
        ],
        'user': {
            'lb': -np.ones(n),
            'ub': np.ones(n),
        },
    }

    alloc_specs = {
        'alloc_f': alloc_f,
        'user': {'m': m, 'num_gens': num_gens},
    }

    # Include @f_i_eval and @df_i_eval if we want to compute gradient in gen
    persis_info['gen_params'].update(
        {
            'L': 1,  # L-smoothness of each function f_i
            'eps': eps,  # error / tolerance
            'rho': rho,
            'N_const': N_const,  # multiplicative constant on numiters
            'step_const': 1,
        }
    )

    if is_manager:
        print('=== Optimizing {} ==='.format(prob_name), flush=True)

    H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs, libE_specs)

    if is_manager:
        print('=== End algorithm ===', flush=True)

        # check we completed
        assert flag == 0

    if is_manager and prob_id <= 4:
        # check we have a Laplacian matrix
        assert la.norm(A.dot(np.zeros(A.shape[1]))) < 1e-15, 'Not a Laplacian matrix'

        # check we have a doubly stochastic matrix
        assert (
            la.norm(np.ones(num_gens) - S.dot(np.ones(num_gens))) / num_gens ** 0.5 < 1e-15
        ), '@S is not a doubly stochastic matrix'

        # compile sum of {f_i} and {x}, and check their values are bounded by O(eps)
        eval_H = H[H['eval_pt']]

        gen_ids = np.unique(eval_H['gen_worker'])
        assert len(gen_ids) == num_gens, 'Gen did not submit any function eval requests'

        x = np.empty(n * num_gens, dtype=float)
        F = 0

        for i, gen_id in enumerate(gen_ids):
            last_eval_idx = np.where(eval_H['gen_worker'] == gen_id)[0][-1]

            f_i = eval_H[last_eval_idx]['f_i']
            x_i = eval_H[last_eval_idx]['x']

            F += f_i
            x[i * n : (i + 1) * n] = x_i

        A_kron_I = spp.kron(A, spp.eye(n))
        consensus_val = np.dot(x, A_kron_I.dot(x))

        assert F - fstar < err_const * eps, 'Error of {:.4e}, expected {:.4e} (assuming f*={:.4e})'.format(
            F - fstar, err_const * eps, fstar
        )
        assert consensus_val < eps, 'Consensus score of {:.4e}, expected {:.4e}\nx={}'.format(consensus_val, eps, x)
